/*_############################################################################
  _##
  _##  SNMP4J-AgentX - BufferedMOMutableTableModel.java
  _##
  _##  Copyright (C) 2005-2014  Frank Fock (SNMP4J.org)
  _##
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  _##  MA  02110-1301  USA
  _##
  _##########################################################################*/

package org.snmp4j.agent.mo;

import org.snmp4j.smi.OID;
import org.snmp4j.smi.Variable;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * The BufferedMOMutableTableModel is a read/write/create table model that dynamically loads data from an
 * backend as needed. The internal buffer holds up to {@link #getMaxBufferSize()} rows fetched from the backend.
 * Rows are removed from the buffer either when they are older than {@link #getBufferTimeoutNanoSeconds()} or
 * if the maximum number of buffered rows is reached.
 * <p>
 * Updates, row creations and removals are directly processed by the backend. Thus, updates are not buffered
 * by this table model.
 * </p>
 *
 * @author Frank Fock
 * @since 2.2
 */
public abstract class BufferedMOMutableTableModel<R extends MOMutableRow2PC>
    extends BufferedMOTableModel<R>
    implements MOMutableTableModel<R>, MOTableRowListener {

  /**
   * Creates a BufferedMOMutableTableModel with the specified {@link org.snmp4j.agent.mo.MOTableRowFactory}.
   *
   * @param rowFactory the row factory to be used to create rows from backend data.
   */
  protected BufferedMOMutableTableModel(MOTableRowFactory<R> rowFactory) {
    super(rowFactory);
  }

  @Override
  public R addRow(R row) {
    Variable[] values = new Variable[row.size()];
    for (int i=0; i<values.length; i++) {
      values[i] = row.getValue(i);
    }
    R previous = getRow(row.getIndex());
    if (previous != null) {
      writeRow(row.getIndex(), values);
    }
    else {
      insertRow(row.getIndex(), values);
    }
    updateBuffer(Collections.singletonList(row), null);
    return previous;
  }

  @Override
  public synchronized void clear() {
    bufferedChunksList.clear();
    bufferedRows.clear();
    removeAllRows();
  }

  @Override
  public synchronized void clear(MOTableRowFilter<R> filter) {
    ArrayList<OID> deleted = new ArrayList<OID>();
    for (Iterator<R> rowIterator = iterator(); rowIterator.hasNext();) {
      R row = rowIterator.next();
      if (!filter.passesFilter(row)) {
        deleted.add(row.getIndex());
      }
    }
    bulkRemoveRows(deleted);
    for (OID index : deleted) {
      BufferedMOTableRow<R> bufferedMOMutableTableRow = bufferedRows.remove(index);
      if (bufferedMOMutableTableRow != null) {
        bufferedChunksList.remove(bufferedMOMutableTableRow);
      }
    }
  }

  protected abstract void removeAllRows();

  protected abstract void bulkRemoveRows(List<OID> indexList);

  protected abstract void insertRow(OID index, Variable[] values);

  protected abstract void writeRow(OID index, Variable[] values);

  protected abstract void deleteRow(OID index);

  @Override
  public void rowChanged(MOTableRowEvent event) {
    if (event.getType() == MOTableRowEvent.UPDATED) {
      Variable[] values = new Variable[event.getRow().size()];
      for (int i=0; i<values.length; i++) {
        values[i] = event.getRow().getValue(i);
      }
      writeRow(event.getRow().getIndex(), values);
    }
  }

  @Override
  public R removeRow(OID index) {
    R previous = getRow(index);
    deleteRow(index);
    return previous;
  }

}
